
/*
 * Program FPGA from configuration image in memory
 *
 * Copyright 2012 Comtrol Corp.
 */

//#define DEBUG
#include <common.h>

#ifdef DEBUG
/* #include <assert.h> */
#define assert(x) do { if (!(x)) { printf("assert fail in %s() at line %d\n",__FUNCTION__,__LINE__); while(1) udelay(1000); } } while(0)
#define xassert(x) /* noop */
#define Log() printf("%s:%d\n",__FUNCTION__,__LINE__)
#else
#define assert(x) ((void)0)
#define Log() /* noop */
#endif


#define PIOA 0xFFFFF400
#define PIOB 0xFFFFF600
#define PIOC 0xFFFFF800

#define PIO_PER 0x0000
#define PIO_PDR 0x0004
#define PIO_PSR 0x0008
#define PIO_OER 0x0010
#define PIO_ODR 0x0014
#define PIO_OSR 0x0018
#define PIO_IFER 0x0020
#define PIO_IFDR 0x0024
#define PIO_IFSR 0x0028
#define PIO_SODR 0x0030
#define PIO_CODR 0x0034
#define PIO_ODSR 0x0038
#define PIO_PDSR 0x003C
#define PIO_IER 0x0040
#define PIO_IDR 0x0044
#define PIO_IMR 0x0048
#define PIO_ISR 0x004C
#define PIO_MDER 0x0050
#define PIO_MDDR 0x0054
#define PIO_MDSR 0x0058
#define PIO_PUER 0x0060
#define PIO_PUDR 0x0064
#define PIO_PUSR 0x0068
#define PIO_ASR 0x0070
#define PIO_BSR 0x0074
#define PIO_ABSR 0x0078
#define PIO_OWER 0x00A0
#define PIO_OWDR 0x00A4
#define PIO_OWSR 0x00A8

#define PMC_PCER 0x0010
#define PMC_PCDR 0x0014
#define PMC_PCSR 0x0018

#define PMC_BASE  0xFFFFFC00
#define PIO_BASE  0xFFFFF400
#define SMC_BASE  0XFFFFEC00

#define ReadByte(addr)  *((volatile uint8_t*)(addr))
#define ReadHWord(addr) *((volatile uint16_t*)(addr))
#define ReadWord(addr)  *((volatile uint32_t*)(addr))

#define WriteByte(addr,value)  (*((volatile uint8_t*)(addr)) = (value))
#define WriteHWord(addr,value) (*((volatile uint16_t*)(addr)) = (value))
#define WriteWord(addr,value)  (*((volatile uint32_t*)(addr)) = (value))

#define PA   (PIO_BASE+0x0000)
#define PB   (PIO_BASE+0x0200)
#define PC   (PIO_BASE+0x0400)

#define SMC0 (SMC_BASE+0x0000)
#define SMC1 (SMC_BASE+0x0010)
#define SMC2 (SMC_BASE+0x0020)
#define SMC3 (SMC_BASE+0x0030)

#define NCS0 0x10000000
#define NCS2 0x30000000

#define Bit(n)  (1<<(n))
#define msleep(ms) udelay((ms)*1000)


#ifdef DEBUG
void showBits(const char *regname, const void *regptr, const char *s0, const char *s1)
{
  unsigned u;
  uint32_t m;
  uint32_t value = ReadWord(regptr);

  printf("  %08lx %s:   %08x   ", (ulong)regptr, regname, value);
  for (u=1,m=0x80000000; m; ++u,m>>=1)
    {
      printf("%1d", !!(value & m));
      if ((u&3)==0)
        printf(" ");
    }
  printf ("\n");

  for (u=0,m=1; m; ++u,m<<=1)
    {
      unsigned b = !!(value & m);
      printf("    %2d  %d  %s\n",u,b,b?s1:s0);
    }
}

void showConfig(const char *name, const void *P)
{
  uint32_t m;
  unsigned b;
  uint32_t PSR  = ReadWord(P+PIO_PSR);
  uint32_t ABSR = ReadWord(P+PIO_ABSR);
  uint32_t OSR  = ReadWord(P+PIO_OSR);
  uint32_t MDSR = ReadWord(P+PIO_MDSR);
  uint32_t OWSR = ReadWord(P+PIO_OWSR);
  uint32_t ODSR = ReadWord(P+PIO_ODSR);
  uint32_t PDSR = ReadWord(P+PIO_PDSR);
  uint32_t PUSR = ReadWord(P+PIO_PUSR);

  printf("%s:\n", name);
  printf(" Bit  Assign  Dir P-up  Tout  Sout  Dout  Din\n");

  for (m=1,b=0; m; ++b,m<<=1)
    {
      printf("  %2d",b);
      printf("  %6s",  (PSR&m) ? "GPIO" : ((ABSR&m) ? "B" : "A"));
      printf("  %3s",  (OSR&m) ? "out" : "in");
      printf("  %3s",  (PUSR&m)? "pup" : "");
      printf("  %4s",  (OSR&m) ? ((MDSR&m) ? "OD" : "bi" ) : "");
      printf("  %4s",  (OSR&m) ? ((OWSR&m) ? "yes" : "no" ) : "");
      printf("  %3d",  !!(ODSR&m));
      printf("  %3d",  !!(PDSR&m));
      printf("\n");
    }
}
#else
#define showBits(regname, regptr, s0, s1)  /* noop */
#define showConfig(name, P)                /* noop */
#endif



void setBits(void *P, uint32_t mask)
{
  WriteWord(P+PIO_SODR, mask);
}

void clrBits(void *P, uint32_t mask)
{
  WriteWord(P+PIO_CODR, mask);
}

uint32_t getBits(void *P, uint32_t mask)
{
  return mask & ReadWord(P+PIO_PDSR);
}


#define SetPROGRAM()  setBits((void *)PB,Bit(16))
#define ClrPROGRAM()  clrBits((void *)PB,Bit(16))
#define GetPROGRAM()  getBits((void *)PB,Bit(16))

#define SetRDWR_B()   setBits((void *)PB,Bit(17))
#define ClrRDWR_B()   clrBits((void *)PB,Bit(17))
#define GetRDWR_B()   getBits((void *)PB,Bit(17))

#define SetCSI_B()    setBits((void *)PB,Bit(18))
#define ClrCSI_B()    clrBits((void *)PB,Bit(18))
#define GetCSI_B()    getBits((void *)PB,Bit(18))

#define GetINIT_B()   getBits((void *)PB,Bit(19))
#define GetDONE()     getBits((void *)PB,Bit(20))

int smc_setup(int ST)
{
  int SV = ST & 0x1f;
  assert((ST & Bit(5)) == 0);
  if (ST >= 128)
    SV |= Bit(5);
  return SV;
}

int smc_pulse(int PT)
{
  int PV = PT & 0x3f;
  assert((PT & Bit(6)) == 0);
  if (PT >= 256)
    PV |= Bit(6);
  return PV;
}

int smc_cycle(int CT)
{
  int CV = CT & 0x7f;
  assert((CT & Bit(7)) == 0);
  CV |= ((CT&0x300)>>1);
  return CV;
}


#define SMC_SETUP  0x00
#define SMC_PULSE  0x04
#define SMC_CYCLE  0x08
#define SMC_MODE   0x0c

void setupSMC0program(void)
{
  WriteWord(SMC0+SMC_SETUP,
            smc_setup( 1) << 24 | // RDCS
            smc_setup( 4) << 16 | // RD
            smc_setup( 1) <<  8 | // WRCS
            smc_setup( 4) <<  0); // WR

  WriteWord(SMC0+SMC_PULSE,
            smc_pulse( 5) << 24 | // RDCS
            smc_pulse( 2) << 16 | // RD
            smc_pulse( 5) <<  8 | // WRCS
            smc_pulse( 2) <<  0); // WR

  WriteWord(SMC0+SMC_CYCLE,
            smc_cycle(9) << 16 | // RD
            smc_cycle(9) <<  0); // WR

  WriteWord(SMC0+SMC_MODE,  (1<<12) + (0<<4) + 0);
}


// SMC MODE register definitions from /usr/src/linux/arch/arm/mach-at91/include/mach/at91sam9_smc.h

#define AT91_SMC_READMODE           (1 << 0)  // Read Mode

#define AT91_SMC_WRITEMODE          (1 << 1)  // Write Mode

#define AT91_SMC_EXNWMODE           (3 << 4)  // NWAIT Mode
#define   AT91_SMC_EXNWMODE_DISABLE (0 << 4)
#define   AT91_SMC_EXNWMODE_FROZEN  (2 << 4)
#define   AT91_SMC_EXNWMODE_READY   (3 << 4)

#define AT91_SMC_BAT                (1 << 8)  // Byte Access Type
#define   AT91_SMC_BAT_SELECT       (0 << 8)
#define   AT91_SMC_BAT_WRITE        (1 << 8)

#define AT91_SMC_DBW                (3 << 12) // Data Bus Width
#define   AT91_SMC_DBW_8            (0 << 12)
#define   AT91_SMC_DBW_16           (1 << 12)
#define   AT91_SMC_DBW_32           (2 << 12)

#define AT91_SMC_TDF              (0xf << 16) // Data Float Time.
#define   AT91_SMC_TDF_(x)        ((x) << 16)

#define AT91_SMC_TDFMODE            (1 << 20) // TDF Optimization - Enabled

#define AT91_SMC_PMEN               (1 << 24) // Page Mode Enabled

#define AT91_SMC_PS                 (3 << 28) // Page Size
#define   AT91_SMC_PS_4             (0 << 28)
#define   AT91_SMC_PS_8             (1 << 28)
#define   AT91_SMC_PS_16            (2 << 28)
#define   AT91_SMC_PS_32            (3 << 28)


static void setupSMC0aiopic(void)
{
  int i;

  WriteWord(SMC0+SMC_SETUP,
            smc_setup( 2) << 24 | // RDCS
            smc_setup( 2) << 16 | // RD
            smc_setup( 2) <<  8 | // WRCS
            smc_setup( 3) <<  0); // WR

  WriteWord(SMC0+SMC_PULSE,
            smc_pulse( 18) << 24 | // RDCS
            smc_pulse( 32) << 16 | // RD
            smc_pulse(  9) <<  8 | // WRCS
            smc_pulse(  8) <<  0); // WR

  WriteWord(SMC0+SMC_CYCLE,
            smc_cycle(39) << 16 | // RD
            smc_cycle(16) <<  0); // WR

  // The select/wait finite-state-machine in the Unity FPGA logic ends
  // up in the wrong state (with WAIT asserted) at the end of the
  // programming process, so do several (one isn't enough) dummy
  // accesses ignoring wait pin to get Unity's select/wait logic back
  // in the correct "idle" state.
  //
  // For some reason if you just read offset 0 a few times, the NAND
  // flash stops working until you read some of the other registers. I
  // don't know why (or which other registers have to be read), but
  // this seems to work...

  // initially, we ignore WAIT pin
  WriteWord(SMC0+SMC_MODE, 0
            //| AT91_SMC_READMODE
            //| AT91_SMC_WRITEMODE
            //| AT91_SMC_EXNWMODE_READY
            | AT91_SMC_EXNWMODE_DISABLE
            | AT91_SMC_BAT_SELECT
            | AT91_SMC_DBW_16
            | AT91_SMC_TDF_(0)
            | AT91_SMC_PS_4);

  for (i=0; i<0x40; i+=2)
    {
      ReadHWord(NCS0+i);
      udelay(1000);
    }

  // now we can pay attention to the WAIT pin
  WriteWord(SMC0+SMC_MODE, 0
            //| AT91_SMC_READMODE
            //| AT91_SMC_WRITEMODE
            | AT91_SMC_EXNWMODE_READY
            //| AT91_SMC_EXNWMODE_DISABLE
            | AT91_SMC_BAT_SELECT
            | AT91_SMC_DBW_16
            | AT91_SMC_TDF_(0)
            | AT91_SMC_PS_4);

  for (i=0; i<0x40; i+=2)
    {
      ReadHWord(NCS0+i);
      udelay(1000);
    }
}


static void setupSMC2rockwell(void)
{
  // set up SMC2 bus timings for Rockell Ethernet Switch

  WriteWord(SMC2+SMC_SETUP,
            smc_setup(  2) << 24 | // RDCS
            smc_setup(  2) << 16 | // RD
            smc_setup(  2) <<  8 | // WRCS
            smc_setup(  2) <<  0); // WR

  WriteWord(SMC2+SMC_PULSE,
            smc_pulse(  10) << 24 | // RDCS
            smc_pulse(  10) << 16 | // RD
            smc_pulse(  10) <<  8 | // WRCS
            smc_pulse(  10) <<  0); // WR

  WriteWord(SMC2+SMC_CYCLE,
            smc_cycle(14)  << 16 | // RD
            smc_cycle(14)  <<  0); // WR

  WriteWord(SMC2+SMC_MODE, 0
            | AT91_SMC_READMODE
            | AT91_SMC_WRITEMODE
            | AT91_SMC_EXNWMODE_DISABLE
            | AT91_SMC_DBW_16
            | AT91_SMC_TDF_(0)
            | AT91_SMC_PS_4);

  // assign PC11 as NCS2
  WriteWord(PC+PIO_ASR, Bit(11));  // select peripheral A
  WriteWord(PC+PIO_PDR, Bit(11));  // select peripheral instead of PIO
  showConfig("PORT C", (void *)PC);
}

void setSMC2timings(unsigned setup, unsigned pulse, unsigned hold)
{
  unsigned cycle = setup+pulse+hold;

  WriteWord(SMC2+SMC_SETUP,
            smc_setup(setup) << 24 | // RDCS
            smc_setup(setup) << 16 | // RD
            smc_setup(setup) <<  8 | // WRCS
            smc_setup(setup) <<  0); // WR

  WriteWord(SMC2+SMC_PULSE,
            smc_pulse(pulse) << 24 | // RDCS
            smc_pulse(pulse) << 16 | // RD
            smc_pulse(pulse) <<  8 | // WRCS
            smc_pulse(pulse) <<  0); // WR

  WriteWord(SMC2+SMC_CYCLE,
            smc_cycle(cycle)  << 16 | // RD
            smc_cycle(cycle)  <<  0); // WR
}

unsigned decode_setup(unsigned v)
{
  return  (128 * ((v>>5)&1)) + (v & 0x1f);
}
unsigned decode_pulse(unsigned v)
{
  return (256 * ((v>>6)&1)) + (v & 0x3f);
}
unsigned decode_cycle(unsigned v)
{
  return (256 * ((v>>7)&3)) + (v & 0x7f);
}

void showSMC2timings(void)
{
  unsigned smc_setup = ReadWord(SMC2+SMC_SETUP);
  unsigned smc_pulse = ReadWord(SMC2+SMC_PULSE);
  unsigned smc_cycle = ReadWord(SMC2+SMC_CYCLE);
  unsigned smc_mode  = ReadWord(SMC2+SMC_MODE);
  unsigned setup,pulse,hold,cycle;


  printf(" SMC_SETUP: %08x\n", smc_setup);
  printf(" SMC_PULSE: %08x\n", smc_pulse);
  printf(" SMC_CYCLE: %08x\n", smc_cycle);
  printf("  SMC_MODE: %08x\n", smc_mode);

  printf("signal setup pulse hold cycle\n");
  
  cycle = decode_cycle((smc_cycle>>16) & 0x1ff);
  setup = decode_setup((smc_setup>>24) & 0x3f);
  pulse = decode_pulse((smc_pulse>>24) & 0x7f);
  hold  = cycle - (setup+pulse);

  printf("  RDCS  %3d  %3d  %3d  %3d\n",setup,pulse,hold,cycle);

  setup = decode_setup((smc_setup>>16) & 0x3f);
  pulse = decode_pulse((smc_pulse>>16) & 0x7f);
  hold  = cycle - (setup+pulse);

  printf("    RD  %3d  %3d  %3d  %3d\n",setup,pulse,hold,cycle);

  cycle = decode_cycle((smc_cycle>> 0) & 0x1ff);
  setup = decode_setup((smc_setup>> 8) & 0x3f);
  pulse = decode_pulse((smc_pulse>> 8) & 0x7f);
  hold  = cycle - (setup+pulse);

  printf("  WRCS  %3d  %3d  %3d  %3d\n",setup,pulse,hold,cycle);

  setup = decode_setup((smc_setup>> 0) & 0x3f);
  pulse = decode_pulse((smc_pulse>> 0) & 0x7f);
  hold  = cycle - (setup+pulse);

  printf("    WR  %3d  %3d  %3d  %3d\n",setup,pulse,hold,cycle);

}


uint8_t bitswapb(uint8_t b)
{
  uint8_t r=0;
  if (b&Bit(0)) r |= Bit(7);
  if (b&Bit(1)) r |= Bit(6);
  if (b&Bit(2)) r |= Bit(5);
  if (b&Bit(3)) r |= Bit(4);
  if (b&Bit(4)) r |= Bit(3);
  if (b&Bit(5)) r |= Bit(2);
  if (b&Bit(6)) r |= Bit(1);
  if (b&Bit(7)) r |= Bit(0);
  return r;
}

uint16_t bitswaph(uint16_t u)
{
  return (bitswapb(u) << 8) | bitswapb(u >> 8);
}



/**
 * verify_fpga_image - verify legacy format FPGA configuration image
 * @img_addr: in RAM address of the legacy format image to be verified
 * @verify: data CRC verification flag
 *
 * verify_fpga_image() verifies legacy image integrity and returns pointer to
 * legacy image header if image verification was completed successfully.
 *
 * returns:
 *     pointer to a legacy image header if valid image was found
 *     otherwise return NULL
 */
static image_header_t *verify_fpga_image(ulong img_addr, int verify)
{
  image_header_t *hdr = (image_header_t *)img_addr;

  if (!image_check_magic(hdr))
    {
      printf("Bad Magic Number\n");
      return NULL;
    }

  if (!image_check_hcrc(hdr))
    {
      printf("Bad Header Checksum\n");
      return NULL;
    }

  if (verify)
    {
      printf("   Verifying Checksum ... ");
      if (!image_check_dcrc (hdr))
        {
          printf ("Bad Data CRC\n");
          return NULL;
        }
      printf ("OK\n");
    }
  return hdr;
}


static void spreadSpectrumClock(int enable)
{
  printf("Selecting %sspread-spectrum clock\n", enable ? "" : "non-");

  if (enable)
    WriteWord(PB+PIO_CODR, Bit(22));  // set output data value to 1
  else
    WriteWord(PB+PIO_SODR, Bit(22));  // set output data value to 0

  // make sure it's configured as GPIO output

  WriteWord(PB+PIO_PER,  Bit(22));  // select PIO function
  WriteWord(PB+PIO_MDDR, Bit(22));  // disable open-drain feature
  WriteWord(PB+PIO_PUDR, Bit(22));  // disable pull-up resister
  WriteWord(PB+PIO_OER,  Bit(22));  // enable as output
}

/*******************************************************************/
/* progfpga - program FPGA from configuration image in memory */
/*******************************************************************/

static int do_progfpga(cmd_tbl_t * cmdtp, int flag, int argc, char *argv[])
{
  image_header_t *hdr;
  ulong img_addr;
  ulong img_data = 0;
  ulong img_len = 0;
  ulong bytecount;
  uint8_t *data;
  int   progOK;
  unsigned i;

  // find out fpga image address
  if (argc < 2)
    {
      img_addr = load_addr;
      debug("*  FPGA: default image load address = 0x%08lx\n", load_addr);
    }
  else
    {
      img_addr = simple_strtoul(argv[1], NULL, 16);
      debug("*  FPGA: cmdline image address = 0x%08lx\n", img_addr);
    }

  // check image type
  if (genimg_get_format((void *)img_addr) != IMAGE_FORMAT_LEGACY)
    {
      printf("Wrong Image Format for %s command\n", cmdtp->name);
      return 1;
    }

  printf("## Found FPGA configuration from Legacy Image at %08lx ...\n", img_addr);
  hdr = verify_fpga_image(img_addr, 1);
  if (!hdr)
    {
      printf("ERROR: invalid FPGA configuration image!\n");
      return 1;
    }
  if (image_get_type(hdr) != IH_TYPE_STANDALONE)
    {
      printf("Wrong Image Type for %s command\n", cmdtp->name);
      return 1;
    }
  // get image data and image length
  img_data = image_get_data (hdr);
  img_len = image_get_data_size (hdr);
  printf("   FPGA configuration data at 0x%08lx, len = 0x%08lx (%ld)\n",
         img_data, img_len, img_len);


  // check Peripheral Clocks
  // Note: the clock of PIO Controller B is disabled after a reset
  if ((ReadWord(PMC_BASE+PMC_PCSR) & Bit(3)) == 0)
    {
      // enable clocks of PIO Controller B
      WriteWord(PMC_BASE+PMC_PCER, Bit(3));
    }
  showBits("PMC_PCSR", (void*)(PMC_BASE+PMC_PCSR), "disabled", "enabled");

  spreadSpectrumClock(0);

  debug("Setup bus timing for SelectMap\n");
  setupSMC0program();

  // outputs on PortB bits 16-18
  WriteWord(PB+PIO_PER,  Bit(16)|Bit(17)|Bit(18));  // select PIO function
  WriteWord(PB+PIO_MDDR, Bit(16)|Bit(17)|Bit(18));  // disable open-drain feature
  WriteWord(PB+PIO_PUDR, Bit(16)|Bit(17)|Bit(18));  // disable pull-up resister
  WriteWord(PB+PIO_SODR, Bit(16)|Bit(17)|Bit(18));  // set output data value to 1
  WriteWord(PB+PIO_OER,  Bit(16)|Bit(17)|Bit(18));  // enable as output

  // inputs on PB bits 19-20
  WriteWord(PB+PIO_PER,  Bit(19)|Bit(20));      // select PIO function
  WriteWord(PB+PIO_MDDR, Bit(19)|Bit(20));      // disable open-drain feature
  WriteWord(PB+PIO_PUDR, Bit(19)|Bit(20));      // disable pull-up resister
  WriteWord(PB+PIO_ODR,  Bit(19)|Bit(20));      // enable as input
  showConfig("PORT B", (void *)PB);

  // initialize
  SetRDWR_B();
  SetCSI_B();

  // reset FPGA
  debug("Resetting FPGA\n");


  ClrPROGRAM();
  msleep(1);
  SetPROGRAM();
  msleep(1);

  debug("Waiting for INIT_B high\n");
  while (GetINIT_B() == 0)
    WriteByte(NCS0, 0);


  debug("Asserting RDWR_B and then CSI_B\n");
  ClrRDWR_B();
  ClrCSI_B();

  printf("Downloading %ld bytes", img_len);

  data = (uint8_t *)img_data;
  bytecount = 0;

#define ENABLE_16BIT_WRITE 1

  while (bytecount < img_len)
    {
      if (ENABLE_16BIT_WRITE && ((bytecount + 2) < img_len))
        {
          WriteHWord(NCS0, bitswaph(*(uint16_t *)&data[bytecount]));
          bytecount += 2;
        }
      else
        {
          WriteByte(NCS0, bitswapb(data[bytecount]));
          ++bytecount;
        }
      if ((bytecount % 10000) == 0)
        printf(".");
      // There is padding at the end of configuration image
      // So check DONE signal when reaching the end of the image
      if (bytecount > (img_len - 24))
        {
          if (GetDONE())
            {
              printf(" DONE\n");
              break;
            }
	  else
            printf("0");
        }
    }

  if (!GetDONE())
    {
      printf("\nEOF and no DONE signal, sending exra clocks: \n");
      for (i=100; i; --i)
        {
          printf("0");
          if (GetDONE())
            {
              printf(" DONE\n");
              break;
            }
          WriteByte(NCS0, 0);
          msleep(1);
        }
    }

  if ( !(progOK = GetDONE()) )
    printf("\nERROR: timed out waiting for DONE\n");
  
  // send extra clocks so startup sequence can complete
  for (i=0; i<1024; ++i)
    WriteByte(NCS0,0);

  printf("Deassert CSI_B then RDWR_B\n");
  SetCSI_B();
  SetRDWR_B();

  if (!progOK)
    return 0;

  //delay to let spread spectrum clock in FPGA stabilize
  msleep(10);

  // According to the docs, the FPGA code we loaded it should be
  // running at this point.  But it isn't.  We have to set up the
  // parallel bus timings and do a couple accesses -- I don't know
  // why.
  printf("Setup bus timing for AIOPIC\n");
  setupSMC0aiopic();

  // Security engine number generator has to be reset by a write
  // to baud rate generator pre scalar register 0xc00. Allow 10
  // useconds for the reset to propagate to the CPLD before sending
  // or receiving any serial data
  printf("Reset security engine number generator\n");
  WriteByte(NCS0+0xc00, 0);

  printf("Setup bus timing for Rockwell Switch\n");
  setupSMC2rockwell();

  spreadSpectrumClock(1);

  return 0;
}

U_BOOT_CMD(progfpga, 2, 0, do_progfpga,
           "program FPGA from configuration image in memory",
           "[address]"
);

static int do_readreg(cmd_tbl_t * cmdtp, int flag, int argc, char *argv[], unsigned base)
{
  ulong offset;
  int size;
  int count = 1;
  int i;

  if ((size = cmd_get_data_size(argv[0], 2)) < 0)
    return 1;

  if (argc < 2)
    {
      printf("Register offset required\n");
      return 1;
    }

  offset = simple_strtoul(argv[1], NULL, 16);

  if (argc > 2)
    count = simple_strtoul(argv[2], NULL, 16);

  for (i=0; i<count; ++i)
    {
      unsigned addr = base+offset+(i*size);
      if ((i*size) % 16 == 0)
        {
          if (i)
            printf("\n");
          printf("%08x:",addr);
        }
      switch (size)
        {
        case 4:
          printf(" %08x",ReadWord(addr));
          break;
        case 2:
          printf(" %04x",ReadHWord(addr));
          break;
        case 1:
          printf(" %02x",ReadByte(addr));
          break;
        }
    }

  printf("\n");
  return 0;
}

static int do_writereg(cmd_tbl_t * cmdtp, int flag, int argc, char *argv[], unsigned base)
{
  ulong offset,value;
  int size;

  if ((size = cmd_get_data_size(argv[0], 2)) < 0)
    return 1;

  if (argc < 3)
    {
      printf("Register offset and value required\n");
      return 1;
    }

  offset = simple_strtoul(argv[1], NULL, 16);
  value  = simple_strtoul(argv[2], NULL, 16);

  switch (size)
    {
    case 4:
      WriteWord(base+offset,value);
      printf("%08lx= %08lx\n",base+offset,value);
      break;
    case 2:
      WriteHWord(base+offset,value);
      printf("%08lx= %04lx\n",base+offset,value & 0xffff);
      break;
    case 1:
      WriteByte(base+offset,value);
      printf("%08lx= %02lx\n",base+offset,value & 0xff);
      break;
    }

  return 0;
}

static int do_writeunity(cmd_tbl_t * cmdtp, int flag, int argc, char *argv[])
{
  return do_writereg(cmdtp, flag, argc, argv, NCS0);
}

static int do_readunity(cmd_tbl_t * cmdtp, int flag, int argc, char *argv[])
{
  return do_readreg(cmdtp, flag, argc, argv, NCS0);
}

U_BOOT_CMD(ru, 3, 1, do_readunity, "read Unity register", "<offset> [count]");
U_BOOT_CMD(wu, 3, 1, do_writeunity, "write Unity register", "<offset> <value>");

static int do_writeswitch(cmd_tbl_t * cmdtp, int flag, int argc, char *argv[])
{
  return do_writereg(cmdtp, flag, argc, argv, NCS2);
}

static int do_readswitch(cmd_tbl_t * cmdtp, int flag, int argc, char *argv[])
{
  return do_readreg(cmdtp, flag, argc, argv, NCS2);
}

U_BOOT_CMD(rs, 3, 0, do_readswitch, "read Rockwell switch register", "<offset> [count]");
U_BOOT_CMD(ws, 3, 0, do_writeswitch, "write Rockwell switch register", "<offset> <value>");

static int testTxControl(int verbose, unsigned loop, unsigned short p)
{
  unsigned short r;
  WriteHWord(NCS0+0x16,p);
  ReadHWord(NCS0+0x14);
  r = ReadHWord(NCS0+0x16);

  if (r==p)
    return 0;

  if (!verbose)
    return 1;

  printf("%6d: exp %04x got %04x",loop,p,r);
  printf(" %04x",ReadHWord(NCS0+0x16));
  printf(" %04x",ReadHWord(NCS0+0x16));
  printf("\n");
  return 1;
}

static int do_unitytest


(cmd_tbl_t * cmdtp, int flag, int argc, char *argv[])
{
  unsigned loops, howmany = 0;
  unsigned i, failures = 0;
  unsigned short p;

  if (argc == 2)
    howmany = simple_strtoul(argv[1], NULL, 10);

  printf("Read/write test on microcode SRAM\n");
  failures = 0;

  for (loops=0; loops<100; ++loops)
    {
      for (p=1; p<31; ++p)
        {
          unsigned char v;
          for (v=p,i=0; i<0x1ff; ++i,++v)
            {
              if (v==0) ++v;
              WriteByte(NCS0+0x400+i, v);
            }
          for (v=p,i=0; i<0x1ff; ++i,++v)
            {
              unsigned char r;
              int j;
              if (v==0) ++v;
              r = ReadByte(NCS0+0x400+i);
              if (r != v)
                {
                  failures += 1;
                  printf("%04x: exp %02x got %02x %02x %02x %02x\n",i,v,r,ReadByte(NCS0+0x400+i),ReadByte(NCS0+0x400+i),ReadByte(NCS0+0x400+i));
                  for (j=0; j<0x400; ++j)
                    {
                      if ((j % 16) == 0)
                        printf("\n%04x",j);
                      printf(" %02x", ReadByte(NCS0+0x400+j));
                    }
                  printf("\n");
                }
            }
        }
    }
  printf("%u passes %u failures\n",loops*30,failures);
  printf("Read/write test on transmit control register\n");
  printf("ffff/0000/5555/AAAA and walking 1/0\n");
  failures = 0;
  failures += testTxControl(1,0,0xffff);
  failures += testTxControl(1,0,0x0000);
  failures += testTxControl(1,0,0x5555);
  failures += testTxControl(1,0,0xaaaa);
  
  for (p = 0x0001; p; p <<= 1)
    {
      failures += testTxControl(1,0,p);
      failures += testTxControl(1,0,~p);
    }

  printf("     %u failures\n", failures);
  if (failures)
    return 1;

  printf("Incrementing pattern\n");

  failures = loops = 0;
  while (loops < 0xffffffff && (howmany == 0  || loops < howmany))
    {
      failures += testTxControl(0,loops,(unsigned short)loops);
      ++loops;
      if ((loops % 100000) == 0)
        {
          if (tstc())
            break;
          if ((loops % 1000000) == 0)
            printf("%5uM loops %u failures\n",loops/1000000,failures);
        }
    }
  
  printf("%u loops %u failures\n",loops,failures);
  return failures != 0;
}

U_BOOT_CMD(unitytest, 2, 0, do_unitytest, "test unity registers", "[count]");


// This file uses different indentation than the rest of the U-Boot
// source files, this is how let emacs know
//
// Local Variables:
// c-file-style: "gnu"
// End:
